#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import collections

__all__ = ["Batch", "BatchIndex", "Offset"]

# Batch = collections.namedtuple(
#    'batch', ['reference', 'positive', 'negative', 'index', 'index_reversed'],
#    defaults=(None, None))


class Batch:
    """A batch of reference, positive, negative samples and an optional index.

    Attributes:
        reference: The reference samples, typically sampled from the prior
            distribution
        positive: The positive samples, typically sampled from the positive
            conditional distribution depending on the reference samples
        negative: The negative samples, typically sampled from the negative
            conditional distribution depending (but often independent) from
            the reference samples
        index: TODO(stes), see docs for multisession training distributions
        index_reversed: TODO(stes), see docs for multisession training distributions
    """

    __slots__ = ["reference", "positive", "negative", "index", "index_reversed"]

    def __init__(self,
                 reference,
                 positive,
                 negative,
                 index=None,
                 index_reversed=None):
        self.reference = reference
        self.positive = positive
        self.negative = negative
        self.index = index
        self.index_reversed = index_reversed

    def to(self, device):
        """Move all batch elements to the GPU."""
        self.reference = self.reference.to(device)
        self.positive = self.positive.to(device)
        self.negative = self.negative.to(device)
        # TODO(stes): Unclear if the indices should also be best represented by
        # torch.Tensors vs. np.ndarrays---this should probably be updated once
        # the GPU implementation of the multi-session sampler is fully ready.
        # if self.index is not None:
        #    self.index = self.index.to(device)
        # if self.index_reversed is not None:
        #    self.index_reversed = self.index_reversed.to(device)


BatchIndex = collections.namedtuple(
    "BatchIndex",
    ["reference", "positive", "negative", "index", "index_reversed"],
    defaults=(None, None),
)


class Offset:
    """Number of samples left and right from an index.

    When indexing datasets, some operations require input of multiple neighbouring samples
    across the time dimension. ``Offset`` represents a simple pair of left and right
    offsets with respect to a index. It provides the range of samples to consider around the current index for
    sampling across the time dimension.

    The provided offsets are positive :py:class:`int`, so that the ``left`` offset corresponds
    to the number of samples to consider previous to the index while the ``right`` offset is strictly positive and
    corresponds to the the index itself and the number of samples to consider following the index.

    Note:
        By convention, the right bound should always be **strictly positive** as it is including the current index itself.
        Hence, for instance, to only consider the current element, you will have to provide (0,1) at :py:class:`Offset` initialization.

    """

    __slots__ = ["left", "right"]

    def __init__(self, *offset):
        if len(offset) == 1:
            (offset,) = offset
            self.left = offset
            self.right = offset
        elif len(offset) == 2:
            self.left, self.right = offset
        else:
            raise ValueError(
                f"Invalid number of elements to bound the Offset, expect 1 or 2 elements, got {len(offset)}."
            )
        self._check_offset_positive()

    def _check_offset_positive(self):
        for offset in [self.right, self.left]:
            if offset < 0:
                raise ValueError(
                    f"Invalid Offset bounds, expect value superior or equal to 0, got {offset}."
                )

        if self.right == 0:
            raise ValueError(
                f"Invalid right bound. By convention, the right bound includes the current index. It should be at least set to 1, "
                f"got {self.right}")

    @property
    def _right(self):
        return None if self.right == 0 else -self.right

    @property
    def left_slice(self):
        """Slice from array start to left border."""
        return slice(0, self.left)

    @property
    def right_slice(self):
        """Slice from right border to array end."""
        return slice(self._right, None)

    @property
    def valid_slice(self):
        """Slice between the two borders."""
        return slice(self.left, self._right)

    def __len__(self):
        return self.left + self.right

    def mask_array(self, array, value):
        array[self.left_slice] = value
        array[self.right_slice] = value
        return array

    def __repr__(self):
        return f"Offset(left = {self.left}, right = {self.right}, length = {len(self)})"



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Joint Allen pseudomouse Ca/Neuropixel datasets.

References:
    * Deitch, Daniel, Alon Rubin, and Yaniv Ziv.
      "Representational drift in the mouse visual cortex."
      Current Biology 31.19 (2021): 4327-4339.
    * de Vries, Saskia EJ, et al.
      "A large-scale standardized physiological survey reveals functional
      organization of the mouse visual cortex."
      Nature Neuroscience 23.1 (2020): 138-151.
    * https://github.com/zivlab/visual_drift
    * http://observatory.brain-map.org/visualcoding
    * https://allensdk.readthedocs.io/en/latest/visual_coding_neuropixels.html
    * Siegle, Joshua H., et al.
      "Survey of spiking in the mouse visual system reveals functional hierarchy."
      Nature 592.7852 (2021): 86-92.
"""

import cebra.data
from cebra.datasets import parametrize
from cebra.datasets.allen import ca_movie
from cebra.datasets.allen import ca_movie_decoding
from cebra.datasets.allen import neuropixel_movie
from cebra.datasets.allen import neuropixel_movie_decoding
from cebra.datasets.allen import NUM_NEURONS
from cebra.datasets.allen import SEEDS
from cebra.datasets.allen import SEEDS_DISJOINT


@parametrize(
    "allen-movie1-ca-neuropixel-10ms-{num_neurons}-{seed}",
    num_neurons=NUM_NEURONS,
    seed=SEEDS,
)
class AllenMovieDataset(cebra.data.DatasetCollection):
    """A joint pseudomouse dataset of 30Hz calcium events and 120 Hz Neuropixels recording during allen Movie1 stimulus.

    It loads instances of AllenCaMovieDataset and AllenNeuropixelMovie120HzDataset for the VISp.

    Args:
        num_neurons: The number of neurons to randomly sample from the stacked pseudomouse neurons. Choose from 10, 30, 50, 100, 200, 400, 600, 800, 900, 1000.
        seed: The random seeds for sampling neurons.

    """

    def __init__(self, num_neurons=1000, seed=111, area="VISp"):
        super().__init__(
            ca_movie.AllenCaMovieDataset(num_neurons, seed, area),
            neuropixel_movie.AllenNeuropixelMovie120HzDataset(
                num_neurons, seed, area),
        )

    def __repr__(self):
        return "CaNeuropixelDataset"


@parametrize(
    "allen-movie-one-ca-neuropixel-{cortex}-{num_neurons}-{split_flag}-10-{seed}",
    cortex=["VISp", "VISpm", "VISam", "VISrl", "VISal", "VISl"],
    num_neurons=NUM_NEURONS,
    split_flag=["train", "test"],
    seed=SEEDS,
)
class AllenMovieOneCaNPCortexDataset(cebra.data.DatasetCollection):
    """A joint pseudomouse dataset of 30Hz calcium events and 120 Hz Neuropixels recording during allen Movie1 stimulus with train/test split.

    It loads instances of AllenCaMoviesDataset and AllenNeuropixelMovie120HzCorticesDataset.

    Args:
        cortex: The cortical area to sample the neurons from.
        num_neurons: The number of neurons to randomly sample from the stacked pseudomouse neurons. Choose from 10, 30, 50, 100, 200, 400, 600, 800, 900, 1000.
        split_flag: The split to load. Choose between `train` and `test`.
        seed: The random seeds for sampling neurons.

    """

    def __init__(self,
                 num_neurons=1000,
                 seed=111,
                 cortex="VISp",
                 split_flag="train"):
        super().__init__(
            ca_movie_decoding.AllenCaMoviesDataset("one", cortex, num_neurons,
                                                   split_flag, seed, 10),
            neuropixel_movie_decoding.
            AllenNeuropixelMovieDecoding120HzCorticesDataset(
                "one", cortex, num_neurons, split_flag, seed),
        )

    def __repr__(self):
        return "CaNeuropixelMovieOneCorticesDataset"


@parametrize(
    "allen-movie-one-ca-neuropixel-{cortex}-disjoint-{group}-{num_neurons}-{split_flag}-10-{seed}",
    cortex=["VISp", "VISam", "VISrl", "VISal"],
    num_neurons=NUM_NEURONS,
    split_flag=["train", "test"],
    group=[0, 1],
    seed=SEEDS_DISJOINT,
)
class AllenMovieOneCaNPCortexDisjointDataset(cebra.data.DatasetCollection):
    """A joint pseudomouse dataset of 30Hz calcium events and 120 Hz Neuropixels recording during allen Movie1 stimulus.

    It loads instances of AllenCaMoviesDisjointDataset and AllenNeuropixelMovie120HzCorticesDisjointDataset.

    Args:
        cortex: The cortical area to sample the neurons from.
        num_neurons: The number of neurons to randomly sample from the stacked pseudomouse neurons. Choose from 10, 30, 50, 100, 200, 400, 600, 800, 900, 1000.
        split_flag: The split to load. Choose between `train` and `test`.
        group: The index of the group among disjoint sets of the sampled neurons.
        seed: The random seeds for sampling neurons.

    """

    def __init__(self, group, num_neurons, seed, cortex, split_flag="train"):
        super().__init__(
            ca_movie_decoding.AllenCaMoviesDisjointDataset(
                "one", cortex, group, num_neurons, split_flag, seed, 10),
            neuropixel_movie_decoding.
            AllenNeuropixelMovie120HzCorticesDisjointDataset(
                group, num_neurons, seed, cortex, split_flag),
        )

    def __repr__(self):
        return "CaNeuropixelMovieOneCorticesDisjointDataset"



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import warnings

import numpy.typing as npt
import packaging
import sklearn
import sklearn.utils.validation as sklearn_utils_validation
import torch

import cebra.helper


def _sklearn_check_array(array, **kwargs):
    # NOTE(stes): See discussion in https://github.com/AdaptiveMotorControlLab/CEBRA/pull/206
    # https://scikit-learn.org/1.6/modules/generated/sklearn.utils.check_array.html
    # force_all_finite was renamed to ensure_all_finite and will be removed in 1.8.
    if packaging.version.parse(
            sklearn.__version__) < packaging.version.parse("1.6"):
        if "ensure_all_finite" in kwargs:
            kwargs["force_all_finite"] = kwargs["ensure_all_finite"]
            del kwargs["ensure_all_finite"]
    return sklearn_utils_validation.check_array(array, **kwargs)


def update_old_param(old: dict, new: dict, kwargs: dict, default) -> tuple:
    """Handle deprecated arguments of a function until they are replaced.

    Note:
        If both the deprecated and new arguments are present, then an error is raised,
        else if only the deprecated argument is present, a warning is raised and the
        old argument is used in place of the new one.

    Args:
        old: A dictionary containing the deprecated arguments.
        new: A dictionary containing the new arguments.
        kwargs: A dictionary containing all the arguments.

    Returns:
        The updated ``kwargs`` set of arguments.

    """
    if kwargs[old] is None and kwargs[new] is None:  # none are present
        kwargs[new] = default
    elif kwargs[old] is not None and kwargs[new] is not None:  # both are present
        raise ValueError(
            f"{old} and {new} cannot be assigned simultaneously. Assign only {new}"
        )
    elif kwargs[old] is not None:  # old version is present but not the new one
        warnings.warn(f"{old} is deprecated. Use {new} instead")
        kwargs[new] = kwargs[old]

    return kwargs


def check_input_array(X: npt.NDArray, *, min_samples: int) -> npt.NDArray:
    """Check validity of the input data, using scikit-learn native function.

    Note:
        * Assert that the array is non-empty, 2D and containing only finite values.
        * Assert that the array has at least {min_samples} samples and 1 feature dimension.
        * Assert that the array is not sparse.
        * Check for the dtype of X and convert values to float if needed.

    Args:
        X: Input data array to check.
        min_samples: Minimum of samples in the dataset.

    Returns:
        The converted and validated array.
    """
    return _sklearn_check_array(
        X,
        accept_sparse=False,
        accept_large_sparse=False,
        dtype=("float16", "float32", "float64"),
        order=None,
        copy=False,
        ensure_2d=True,
        ensure_all_finite=True,
        allow_nd=False,
        ensure_min_samples=min_samples,
        ensure_min_features=1,
    )


def check_label_array(y: npt.NDArray, *, min_samples: int):
    """Check validity of the labels, using scikit-learn native function.

    Note:
        * Assert that the array is non-empty and containing only finite values.
        * Assert that the array has at least {min_samples} samples.
        * Assert that the array is not sparse.
        * Check for the dtype of y and convert values to numeric if needed.

    Args:
        y: Labels array to check.
        min_samples: Minimum of samples in the label array.

    Returns:
        The converted and validated labels.
    """
    return _sklearn_check_array(
        y,
        accept_sparse=False,
        accept_large_sparse=False,
        dtype="numeric",
        order=None,
        copy=False,
        ensure_2d=False,
        ensure_all_finite=True,
        allow_nd=False,
        ensure_min_samples=min_samples,
    )


def check_device(device: str) -> str:
    """Select a device depending on the requirement and availabilities.

    Args:
        device: The device to return, if possible.

    Returns:
        Either cuda, cuda:device_id, mps, or cpu depending on {device} and availability in the environment.
    """

    if device == "cuda_if_available":
        if torch.cuda.is_available():
            return "cuda"
        elif cebra.helper._is_mps_availabe(torch):
            return "mps"
        else:
            return "cpu"
    elif device.startswith("cuda:") and len(device) > 5:
        cuda_device_id = device[5:]
        if cuda_device_id.isdigit():
            device_count = torch.cuda.device_count()
            device_id = int(cuda_device_id)
            if device_id < device_count:
                return f"cuda:{device_id}"
            else:
                raise ValueError(
                    f"CUDA device {device_id} is not available. Available device IDs are 0 to {device_count - 1}."
                )
        else:
            raise ValueError(
                f"Invalid CUDA device ID format. Please use 'cuda:device_id' where '{cuda_device_id}' is an integer."
            )
    elif device == "cuda" and torch.cuda.is_available():
        return "cuda:0"
    elif device == "cpu":
        return device
    elif device == "mps":
        if not torch.backends.mps.is_available():
            if not torch.backends.mps.is_built():
                raise ValueError(
                    "MPS not available because the current PyTorch install was not "
                    "built with MPS enabled.")
            else:
                raise ValueError(
                    "MPS not available because the current MacOS version is not 12.3+ "
                    "and/or you do not have an MPS-enabled device on this machine."
                )

        return device

    raise ValueError(f"Device needs to be cuda, cpu or mps, but got {device}.")


def check_fitted(model: "cebra.models.Model") -> bool:
    """Check if an estimator is fitted.

    Args:
        model: The model to assess.

    Returns:
        True if fitted.
    """
    return hasattr(model, "n_features_")



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A simple API for loading various data formats used with CEBRA.

Availability of different data formats depends on the installed
dependencies. If a dependency is not installed, an attempt to load
a file of that format will throw an error with further installation
instructions.

Currently available formats:

- HDF5 via ``h5py``
- Pickle files via ``pickle``
- Joblib files via ``joblib``
- Various dataframe formats via ``pandas``.
- Matlab files via ``scipy.io.loadmat``
- DeepLabCut (single animal) files via ``deeplabcut``
"""

import abc
import pathlib
import warnings
from typing import IO, List, Optional, Union

import numpy as np
import numpy.typing as npt
import torch

_IS_H5PY_AVAILABLE = True
try:
    import h5py
except ModuleNotFoundError:
    _IS_H5PY_AVAILABLE = False
    warnings.warn(
        "h5py module was not found, be sure it is installed in your env.",
        ImportWarning)

_IS_PANDAS_AVAILABLE = True
try:
    import pandas as pd
except ModuleNotFoundError:
    _IS_PANDAS_AVAILABLE = False
    warnings.warn(
        "pandas module was not found, be sure it is installed in your env.",
        ImportWarning,
    )

_IS_JOBLIB_AVAILABLE = True
try:
    import joblib as jl
except ModuleNotFoundError:
    _IS_JOBLIB_AVAILABLE = False
    warnings.warn(
        "joblib module was not found, be sure it is installed in your env.",
        ImportWarning,
    )

_IS_PICKLE_AVAILABLE = True
try:
    import pickle
except ModuleNotFoundError:
    _IS_PICKLE_AVAILABLE = False
    warnings.warn(
        "pickle module was not found, be sure it is installed in your env.",
        ImportWarning,
    )

_IS_SCIPY_AVAILABLE = True
try:
    import scipy.io
except ModuleNotFoundError:
    _IS_SCIPY_AVAILABLE = False
    warnings.warn(
        "scipy module was not found, be sure it is installed in your env.",
        ImportWarning,
    )

_IS_DLC_INTEGRATION_AVAILABLE = True
try:
    if _IS_PANDAS_AVAILABLE:
        from cebra.integrations.deeplabcut import load_deeplabcut
except ModuleNotFoundError:
    _IS_DLC_INTEGRATION_AVAILABLE = False
    warnings.warn(
        "DLC integration module was not found, be sure it is installed in your env.",
        ImportWarning,
    )


def _module_not_found_error(module_name):
    return ModuleNotFoundError(
        f"Could not load {module_name}. You can manually install {module_name} "
        "or install the [datasets] dependency in cebra: "
        "pip install 'cebra[datasets]'")


class _BaseLoader(abc.ABC):
    """Base loader."""

    @abc.abstractmethod
    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        """Load data from file, at key.

        Args:
            file: The path to the given file to load, in a supported format.
            key: The key referencing the data of interest in the file, if the file has a dictionary-like structure.
            columns: The part of the data to keep in the output 2D-array. For now, it corresponds to the columns of
                a DataFrame to keep if the data selected is a DataFrame.

        Returns:
            The loaded data.
        """
        raise NotImplementedError()


class _NumpyLoader(_BaseLoader):
    """Loader for numpy files.

    Supports ``.npy``.
    """

    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        try:
            loaded_array = np.load(file)
        except ValueError:
            loaded_array = np.load(file, allow_pickle=True)
        return loaded_array


class _NumpyZipLoader(_NumpyLoader):
    """Loader for zipped numpy files.

    Supports ``.npz``.
    """

    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        loaded_zip = _NumpyLoader.load(file, key)
        if key is not None:
            if key in loaded_zip.files:
                if (isinstance(loaded_zip[key], np.ndarray) and
                        "str" not in loaded_zip[key].dtype.name
                   ):  # check that key is valid
                    loaded_array = loaded_zip[key]
                else:
                    raise AttributeError(
                        f"key={key} does not correspond to a valid np.array field of your .npz file."
                    )
            else:
                raise AttributeError(
                    f"key={key} is not a field of your .npz file.")
        else:  # take the first array of the dict
            found_array = False
            for key in list(loaded_zip.keys()):
                if "str" not in loaded_zip[key].dtype.name:
                    loaded_array = loaded_zip[key]
                    found_array = True
                    break
            if not found_array:
                raise AttributeError("No valid array was found in your file.")
        return loaded_array


class _H5pyLoader(_BaseLoader):
    """Loader for HDF5 files.

    Supports ``.h5``, ``.h``, ``.hdf``, ``.hdf5`` as well as the .h5 output files from DLC.
    """

    def load(
        filename: Union[str, pathlib.Path],
        key: Optional[Union[str, int]],
        columns: Optional[list],
    ) -> npt.NDArray:
        if _IS_H5PY_AVAILABLE:
            with h5py.File(filename, "r") as loaded_h5_file:
                # get keys in the file
                df_keys, array_keys = _H5pyLoader._get_keys(loaded_h5_file, key)

                # if the file contains pd.DataFrame and
                # the required key is not an array in the file
                if df_keys and key not in array_keys:
                    # if the pd.DataFrame is a DLC prediction output
                    if key is None and _H5pyLoader._is_dlc_df(
                            filename, df_keys):
                        if _IS_DLC_INTEGRATION_AVAILABLE:
                            loaded_array = load_deeplabcut(filepath=filename,
                                                           keypoints=columns)
                        else:
                            raise ModuleNotFoundError(
                                "DLC integration could not be loaded. "
                                "Most likely, this is because you do not have all "
                                "integrations dependencies installed. Try installing "
                                "cebra with the [integrations] and [datasets] dependency to fix this "
                                "error. You might need to re-start your environment "
                                "after installing: "
                                "pip install 'cebra[integrations,datasets]'.")
                    # if the provided key is valid
                    elif key in df_keys:
                        loaded_array = _PandasLoader.load_from_h5(
                            filename, key, columns)
                    # if no key provided but only one data structure in the file
                    elif len(df_keys) == 1 and not array_keys:
                        loaded_array = _PandasLoader.load_from_h5(
                            filename, df_keys[0], columns)
                    else:
                        raise AttributeError(
                            f"Data in a pd.DataFrame format, please provide the corresponding key, "
                            f"we expect a value from {df_keys}.")
                # if there is no pd.DataFrame or the key corresponds to a numpy.array
                # it will either take the data structure associated with ``key`` or take
                # the first numpy.array found in the data structure
                elif array_keys:
                    loaded_array = loaded_h5_file.get(array_keys[0])[:]
                else:
                    raise AttributeError(
                        "No valid data structure was found in your file.")
        else:
            raise _module_not_found_error("h5py")
        return loaded_array

    @staticmethod
    def _is_df(h5_file: IO[bytes], key: str) -> bool:
        """Check if the ``.h5`` file ``h5_file`` has a format valid to contain a :py:class:`pandas.DataFrame`.

        Args:
            h5_file: A ``.h5`` file to get the keys from.
            key: The key to fetch if it is contained in ``h5_file``.

        Returns:
            True if ``h5_file`` has a valid format to contain a :py:class:`pandas.DataFrame`.
        """
        return isinstance(h5_file.get(key), h5py._hl.group.Group)

    @staticmethod
    def _is_dlc_df(h5_file: IO[bytes], df_keys: List[str]) -> bool:
        """Check if the ``.h5`` file ``h5_file`` has a :py:class:`pandas.DataFrame` table format corresponding
        to the DLC output prediction file format.

        Args:
            h5_file: A ``.h5`` file to check the format.

        Returns:
            True if ``h5_file`` has a valid format to be a DLC output prediction file format.
        """
        try:
            if ["_i_table", "table"] in df_keys:
                df = read_hdf(h5_file, key="table")
            else:
                df = read_hdf(h5_file, key=df_keys[0])
        except KeyError:
            df = read_hdf(h5_file)
        return all(value in df.columns.names
                   for value in ["scorer", "bodyparts", "coords"])

    @staticmethod
    def _is_2D_array(h5_file: IO[bytes], key: str) -> bool:
        """Check if the ``.h5`` file ``h5_file`` has a format valid to contain a :py:func:`numpy.array`.

        Args:
            h5_file: A ``.h5`` file to get the keys from.
            key: The key to fetch if it is contained in ``h5_file``.

        Returns:
            True if ``h5_file`` has a valid format to contain a :py:func:`numpy.array`.
        """
        return (h5_file.get(key).dtype != "object" and
                "u" not in h5_file.get(key).dtype.name)

    @staticmethod
    def _get_keys(h5_file: IO[bytes], key: Optional[str]) -> List[str]:
        """Returns either all the keys contained in the ``.h5`` file ``h5_file`` or ``key`` if contained in ``h5_file``.

        Args:
            h5_file: A ``.h5`` file to get the keys from.
            key: The key to return if it is contained in ``h5_file``.

        Returns:
            A list containing all keys or the key provided as parameter or is empty if the provided key is
            not contained in the :py:class:`pandas.DataFrame`.
        """
        df_keys = []
        array_keys = []
        if key is None:
            for k in list(h5_file.keys()):
                if _H5pyLoader._is_df(h5_file, k):
                    df_keys.append(k)
                elif _H5pyLoader._is_2D_array(h5_file, k):
                    array_keys.append(k)
        elif _H5pyLoader._is_df(h5_file, key):
            df_keys.append(key)
        elif _H5pyLoader._is_2D_array(h5_file, key):
            array_keys.append(key)
        else:
            raise AttributeError(
                f"key={key} does not correspond to a valid np.array or pd.DataFrame field of your .h5 file."
                f"Make sure that your key is valid, got {key}")
        return df_keys, array_keys


class _PandasLoader(_BaseLoader):
    """Loader for files containing :py:class:`pandas.DataFrame`."""

    @staticmethod
    def load_from_h5(file: Union[pathlib.Path, str], key: str,
                     columns: Optional[List[str]]) -> npt.NDArray:
        """Load data stored in a :py:class:`pandas.DataFrame` from the ``.h5`` ``file`` provided.

        Args:
            file: The file to fetch in.
            key: The key associated to the data of interest in the ``file``.
            columns: The columns of the :py:class:`pandas.DataFrame` to keep as the data of interest.

        Note:
            If ``columns`` are provided, the :py:class:pandas.DataFrame` should contain a unique level of
            column indexes.

        Returns:
            A :py:func:`numpy.array` containing the data of interest extracted from the :py:class:`pandas.DataFrame`.
        """
        df = read_hdf(file, key=key)
        if columns is None:
            loaded_array = df.values
        elif isinstance(columns, list) and df.columns.nlevels == 1:
            # Check the column validity
            for c in columns:
                if c not in list(df.columns):
                    raise AttributeError(
                        f"{c} is not a valid column of the dataframe contained in your file, expected values from {list(df.columns)}."
                    )
            loaded_array = df[columns].values
        else:
            raise AttributeError(
                f"DataFrame loading is only handled for dataframes with a single column index, got {df.columns.nlevels} columns indexes."
            )
        return loaded_array


class _PyTorchLoader(_BaseLoader):
    """Loader for PyTorch files.

    Supports ``.pt``, ``.p``.
    """

    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        torch_file = torch.load(file)
        if torch.is_tensor(torch_file):
            loaded_array = torch_file.numpy()
        elif type(torch_file) is dict:
            if key is not None:
                if key in torch_file.keys() and torch.is_tensor(
                        torch_file.get(key)[:]):  # check that key is valid
                    loaded_array = torch_file.get(key)[:].numpy()
                else:
                    raise AttributeError(
                        f"key={key} does not correspond to a valid field of your .pt file."
                    )
            else:
                found_array = False
                for key in list(torch_file.keys()):
                    if torch.is_tensor(torch_file.get(key)[:]):
                        loaded_array = torch_file.get(key)[:].numpy()
                        found_array = True
                        break
                if not found_array:
                    raise AttributeError(
                        "No valid array was found in your file.")
        else:
            raise AttributeError(
                f"{file} contains a model. Be sure to provide a tensor or a dict."
            )
        return loaded_array


class _CsvLoader(_BaseLoader):
    """Loader for CSV files.

    Supports ``.csv``.
    """

    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        if _IS_PANDAS_AVAILABLE:
            try:
                loaded_array = pd.read_csv(file, sep=",", header=None).values
            except pd.errors.EmptyDataError:
                raise AttributeError(".csv file is empty.")
        else:
            raise _module_not_found_error("pandas")
        return loaded_array


class _ExcelLoader(_BaseLoader):
    """Loader for Excel files.

    Supports ``.xls``, ``.xlsx``, ``.xlsm``.
    """

    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        if _IS_PANDAS_AVAILABLE:
            loaded_dict = pd.read_excel(file, header=None,
                                        sheet_name=None)  # dict of df
            if (len(list(loaded_dict.keys())) == 1 and
                    "Sheet" in loaded_dict.keys() and
                    loaded_dict["Sheet"].empty):
                raise AttributeError(".xls file is empty.")

            if key is not None:
                if key in loaded_dict.keys():
                    loaded_array = loaded_dict[key].values
                else:
                    raise AttributeError(
                        f"key={key} is not a valid name for a sheet of your excel file."
                    )
            else:  # take the first sheet of the excel file, we know not empty already
                for key in loaded_dict.keys():
                    loaded_array = loaded_dict[key].values
                    break
        else:
            raise _module_not_found_error("pandas")
        return loaded_array

    # def prepare_engine(extension: str):
    #     if extension in [".xls", ".xlsx", ".xlsm"]:
    #         engine = "xlrd"
    #     elif extension == ".xlsb":
    #         engine = "pyxlsb"
    #     elif extension in [".odf", ".ods", ".odt"]:
    #         engine = "odf"
    #     else:
    #         raise AttributeError(
    #             f"{extension} endfile is not a valid extension for excel files."
    #         )
    #     return engine


class _JoblibLoader(_BaseLoader):
    """Loader for JobLib files.

    Supports ``.jl``.
    """

    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        if _IS_JOBLIB_AVAILABLE:
            loaded_data = jl.load(file, mmap_mode="r")

            # NOTE: data will be stored in a np.array upon loading,
            # if originally np.array, then it is directly it
            # else if dict, it will be of format [dict()]
            # so need to do loaded_data[0] to access the dict
            if type(loaded_data) is dict:  # it is in a dict format: [dict]
                if key is not None:
                    if key in loaded_data.keys():
                        if (isinstance(loaded_data[key], np.ndarray) and
                                "str" not in loaded_data[key].dtype.name
                           ):  # check that key is valid
                            loaded_array = loaded_data[key]
                        else:
                            raise AttributeError(
                                f"key={key} does not correspond to a valid field in the .jl file."
                            )
                    else:
                        raise AttributeError(
                            f"key={key} is not a field in your .jl file.")
                else:
                    found_array = False
                    for key in loaded_data.keys():
                        if "str" not in str(type(loaded_data[key])):
                            loaded_array = loaded_data[key]
                            found_array = True
                            break
                    if not found_array:
                        raise AttributeError(
                            "No valid array was found in your file.")
            elif type(np.array(loaded_data.tolist)) is np.ndarray:
                loaded_array = loaded_data
            else:
                # TODO: make it more robust to other types of data in the .jl file
                raise NotImplementedError(
                    f"{type(loaded_data)} is not handled for .jl files.")
        else:
            raise _module_not_found_error("joblib")
        return loaded_array


class _PickleLoader(_BaseLoader):
    """Loader for pickle files.

    Supports ``.pkl``, ``.p``.
    """

    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        if _IS_PICKLE_AVAILABLE:
            with open(file, "rb") as pk:
                loaded_data = pickle.load(pk)
            if isinstance(loaded_data, np.ndarray):
                loaded_array = loaded_data
            elif type(loaded_data) is dict:
                if key is not None:
                    if (key in loaded_data.keys() and type(loaded_data[key])
                            is np.ndarray):  # check that key is valid
                        loaded_array = loaded_data[key]
                    else:
                        raise AttributeError(
                            f"key={key} does not correspond to a valid np.array field of the dict stored in your .pk file"
                        )
                else:
                    found_array = False
                    for key in loaded_data.keys():
                        if type(loaded_data[key]) is np.ndarray:
                            loaded_array = loaded_data[key]
                            found_array = True
                            break
                    if not found_array:
                        raise AttributeError(
                            "No valid array was found in your file.")

            else:
                # TODO: make it more robust to other types of data in the .pk file
                raise NotImplementedError(
                    f"{type(loaded_data)} is not handled for .pk files.")
        else:
            raise _module_not_found_error("pickle")
        return loaded_array


class _MatfileLoader(_BaseLoader):
    """Loader for MAT-files.

    Supports ``.mat``. Newer and older versions.
    """

    def load(
        file: Union[str, pathlib.Path],
        key: Optional[Union[str, int]] = None,
        columns: Optional[list] = None,
    ) -> npt.NDArray:
        try:
            loaded_mat = scipy.io.loadmat(file)
            if key is not None:
                if (key in loaded_mat.keys() and
                        type(loaded_mat[key]) is np.ndarray and
                        "str" not in loaded_mat[key].dtype.name):
                    loaded_array = loaded_mat[key]
                else:
                    raise AttributeError(
                        f"key={key} does not correspond to a valid np.array field of the dict stored in your .mat file"
                    )
            else:
                found_array = False
                for key in loaded_mat.keys():
                    if (type(loaded_mat[key]) is np.ndarray and
                            "str" not in loaded_mat[key].dtype.name):
                        loaded_array = loaded_mat[key]
                        found_array = True
                        break
                if not found_array:
                    raise AttributeError(
                        "No valid array was found in your file.")
        except NotImplementedError:
            if _IS_H5PY_AVAILABLE:
                loaded_array = _H5pyLoader.load(file, key)
            else:
                raise _module_not_found_error("h5py")
        return loaded_array


# loaders associating every handled file ending to corresponding class
__loaders = {
    ".npy": _NumpyLoader,
    ".npz": _NumpyZipLoader,
    ".h5": _H5pyLoader,
    ".h": _H5pyLoader,
    ".hdf": _H5pyLoader,
    ".hdf5": _H5pyLoader,
    ".pt": _PyTorchLoader,
    ".pth": _PyTorchLoader,
    ".csv": _CsvLoader,
    ".xls": _ExcelLoader,
    ".xlsx": _ExcelLoader,
    ".xlsm": _ExcelLoader,
    # ".xlsb": _ExcelLoader,
    # ".odf": _ExcelLoader,
    # ".ods": _ExcelLoader,
    # ".odt": _ExcelLoader,
    ".jl": _JoblibLoader,
    ".pkl": _PickleLoader,
    ".p": _PickleLoader,
    ".mat": _MatfileLoader,
}


def load(
    file: Union[str, pathlib.Path],
    key: Optional[Union[str, int]] = None,
    columns: Optional[list] = None,
) -> npt.NDArray:
    """Load a dataset from the given file.

    The following file types are supported:
        - Numpy files: npy, npz;
        - HDF5 files: h5, hdf, hdf5, including h5 generated through DLC;
        - PyTorch files: pt, p;
        - csv files;
        - Excel files: xls, xlsx, xlsm;
        - Joblib files: jl;
        - Pickle files: p, pkl;
        - MAT-files: mat.

    The assumptions on your data are as following:
        - it contains at least one data structure (e.g. a numpy array, a torch.Tensor, etc.);
        - it can be directly in the form of a collection (e.g. a dictionary);
        - if the file contains a collection, the user can provide a key to refer to the data value they want to access;
        - if no key is provided, the first data structure found upon iteration of the collection will be loaded;
        - if a key is provided, it needs to correspond to an existing item of the collection;
        - if a key is provided, the data value accessed needs to be a data structure;
        - the function loads data for only one data structure, even if the file contains more. The function can be
          called again with the corresponding key to get the other ones.

    Args:
        file: The path to the given file to load, in a supported format.
        key: The key referencing the data of interest in the file, if the file has a dictionary-like structure.
        columns: The part of the data to keep in the output 2D-array. For now, it corresponds to the columns of
            a DataFrame to keep if the data selected is a DataFrame.


    Returns:
        The loaded data.

    Example:

        >>> import cebra
        >>> import cebra.helper as cebra_helper
        >>> import numpy as np
        >>> # Create the files to load the data from
        >>> # Create a .npz file
        >>> X = np.random.normal(0,1,(100,3))
        >>> y = np.random.normal(0,1,(100,4))
        >>> np.savez("data", neural = X, trial = y)
        >>> # Create a .h5 file
        >>> url = "https://github.com/DeepLabCut/DeepLabCut/blob/main/examples/Reaching-Mackenzie-2018-08-30/labeled-data/reachingvideo1/CollectedData_Mackenzie.h5?raw=true"
        >>> dlc_file = cebra_helper.download_file_from_url(url) # an .h5 example file
        >>> # Load data
        >>> X = cebra.load_data(file="data.npz", key="neural")
        >>> y_trial_id = cebra.load_data(file="data.npz", key="trial")
        >>> y_behavior = cebra.load_data(file=dlc_file, columns=["Hand", "Tongue"])

    """
    file_ending = pathlib.Path(file).suffix
    loader = _get_loader(file_ending)
    data = loader.load(file, key=key, columns=columns)
    return data


def _get_loader(file_ending: str) -> _BaseLoader:
    """Get corresponding class based on handled file ending.

    Args:
        file_ending: The file ending of the handled file.

    Raises:
        OSError: The file ending is not supported.

    Returns:
        The loader class corresponding to the handle file.
    """
    if file_ending not in __loaders.keys() or file_ending == "":
        raise OSError(f"File ending {file_ending} not supported.")
    return __loaders[file_ending]


def read_hdf(filename, key=None):
    """Read HDF5 file using pandas, with fallback to h5py if pandas fails.

    Args:
        filename: Path to HDF5 file
        key: Optional key to read from HDF5 file. If None, tries "df_with_missing"
             then falls back to first available key.

    Returns:
        pandas.DataFrame: The loaded data

    Raises:
        RuntimeError: If both pandas and h5py fail to load the file
    """

    return pd.read_hdf(filename, key=key)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools

import pytest
import torch
from torch import nn

import cebra.models
import cebra.models.model
import cebra.registry


def test_registry():
    assert cebra.registry.is_registry(cebra.models)
    assert cebra.registry.is_registry(cebra.models, check_docs=True)


@pytest.mark.parametrize(
    "model_name,batch_size,input_length",
    list(itertools.product(cebra.models.get_options(), [1, 2, 7], [100, 430])),
)
def test_offset_models(model_name, batch_size, input_length):
    model = cebra.models.init(model_name,
                              num_neurons=5,
                              num_output=3,
                              num_units=4)
    assert isinstance(model, cebra.models.Model)
    assert isinstance(model, cebra.models.HasFeatureEncoder)

    if isinstance(model, cebra.models.ClassifierModel):
        model.set_output_num(label_num=12, override=False)

    offset = model.get_offset()

    # batched input
    inputs = torch.randn((batch_size, 5, len(offset)))
    outputs = model.net(inputs)
    assert isinstance(outputs, torch.Tensor)
    assert outputs.shape == (batch_size, 3)

    # check that full model works as well
    outputs = model(inputs)
    if isinstance(outputs, tuple):
        for output in outputs:
            assert isinstance(output, torch.Tensor)
            assert len(output) == batch_size
    else:
        assert isinstance(outputs, torch.Tensor)
        assert len(outputs) == batch_size

    # full length input
    if isinstance(model, cebra.models.ConvolutionalModelMixin):
        inputs = torch.randn((batch_size, 5, input_length))
        outputs = model.net(inputs)
        if isinstance(model, cebra.models.ResampleModelMixin):
            assert outputs.shape == (
                batch_size,
                3,
                (input_length - len(offset)) // model.resample_factor + 1,
            )
        else:
            assert outputs.shape == (batch_size, 3,
                                     input_length - len(offset) + 1)

        # check that full model works as well
        outputs = model(inputs)
        assert isinstance(outputs, torch.Tensor)
        assert len(outputs) == batch_size


def test_multiobjective():

    class TestModel(cebra.models.Model):

        def __init__(self):
            super().__init__(num_input=10, num_output=10)
            self._model = nn.Linear(self.num_input, self.num_output)

        def forward(self, x):
            return self._model(x)

        @property
        def get_offset(self):
            return None

    model = TestModel()

    multi_model_overlap = cebra.models.MultiobjectiveModel(
        model,
        dimensions=(4, 6),
        output_mode="overlapping",
        append_last_dimension=True)
    multi_model_separate = cebra.models.MultiobjectiveModel(
        model,
        dimensions=(4, 6),
        output_mode="separate",
        append_last_dimension=True)

    x = torch.randn(5, 10)

    assert model(x).shape == (5, 10)

    assert model.num_output == multi_model_overlap.num_output
    assert model.get_offset == multi_model_overlap.get_offset

    first, second, third = multi_model_overlap(x)
    assert first.shape == (5, 4)
    assert second.shape == (5, 6)
    assert third.shape == (5, 10)

    first, second, third = multi_model_separate(x)
    assert first.shape == (5, 4)
    assert second.shape == (5, 2)
    assert third.shape == (5, 4)


@pytest.mark.parametrize("version,raises", [
    ["1.12", False],
    ["2.", False],
    ["2.0.0rc", False],
    ["2.0", False],
    ["2.5", False],
    ["1.11.0rc1", True],
    ["1.10", True],
    ["1.2", True],
    ["1.0", True],
])
def test_version_check(version, raises):

    torch.__version__ = version
    assert cebra.models.model._check_torch_version(
        raise_error=False) == (not raises)
    if raises:
        with pytest.raises(ImportError):
            cebra.models.model._check_torch_version(raise_error=True)


def test_version_check_dropout_available():
    raises = cebra.models.model._check_torch_version(raise_error=False)
    if raises:
        assert len(cebra.models.get_options("*dropout*")) == 0
    else:
        assert len(cebra.models.get_options("*dropout*")) > 0



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Wrappers for using models with multiobjective solvers.

.. note::

    Experimental as of Nov 06, 2022.
"""

from typing import Tuple

import torch
from torch import nn

import cebra.models


class _Norm(nn.Module):
    """Normalize the input tensor across its first dimension.

    TODO:
        * Move this class to ``cebra.models.layers``.
    """

    def forward(self, inp):
        """Normalize the input tensor across its first dimension."""
        return inp / torch.norm(inp, dim=1, keepdim=True)


class MultiobjectiveModel(nn.Module):
    """Wrapper around contrastive learning models to all training with multiple objectives

    Multi-objective training splits the last layer's feature representation into multiple
    chunks, which are then used for individual training objectives.

    Args:
        module: The module to wrap
        dimensions: A tuple of dimension values to extract from the model's feature embedding.
        renormalize: If True, the individual feature slices will be re-normalized before
            getting returned---this option only makes sense in conjunction with a loss based
            on the cosine distance or dot product.
        output_mode: A mode as defined in ``MultiobjectiveModel.Mode``. Overlapping means that
            when ``dimensions`` are set to `(x0, x1, ...)``, features will be extracted from
            ``0:x0, 0:x1, ...``. When mode is set to separate, features are extracted from
            ``x0:x1, x1:x2, ...``.
        append_last_dimension: Defaults to True, and will allow to omit the last dimension in
            the ``dimensions`` argument (which should be equal to the output dimension) of the
            given model.

    TODO:
        - Update nn.Module type annotation for ``module`` to cebra.models.Model
    """

    class Mode:
        """Mode for slicing and potentially normalizing the output embedding.

        The options are:

        - ``OVERLAPPING``: When ``dimensions`` are set to `(x0, x1, ...)``, features will be
          extracted from ``0:x0, 0:x1, ...``.
        - ``SEPARATE``: Features are extracted from ``x0:x1, x1:x2, ...``

        """

        OVERLAPPING = "overlapping"
        SEPARATE = "separate"
        _ALL = {OVERLAPPING, SEPARATE}

        def is_valid(self, mode):
            """Check if a given string representation is valid.

            Args:
                mode: String representation of the mode.

            Returns:
                ``True`` for a valid representation, ``False`` otherwise.
            """
            return mode in _ALL  # noqa: F821

    def __init__(
        self,
        module: nn.Module,
        dimensions: Tuple[int],
        renormalize: bool = False,
        output_mode: str = "overlapping",
        append_last_dimension: bool = False,
    ):
        super().__init__()

        if not isinstance(module, cebra.models.Model):
            raise ValueError("Can only wrap models that are subclassing the "
                             "cebra.models.Model abstract base class. "
                             f"Got a model of type {type(module)}.")

        self.module = module
        self.renormalize = renormalize
        self.output_mode = output_mode

        self._norm = _Norm()
        self._compute_slices(dimensions, append_last_dimension)

    @property
    def get_offset(self):
        """See :py:meth:`cebra.models.model.Model.get_offset`."""
        return self.module.get_offset

    @property
    def num_output(self):
        """See :py:attr:`cebra.models.model.Model.num_output`."""
        return self.module.num_output

    def _compute_slices(self, dimensions, append_last_dimension):

        def _valid_dimensions(dimensions):
            return max(dimensions) == self.num_output

        if append_last_dimension:
            if _valid_dimensions(dimensions):
                raise ValueError(
                    f"append_last_dimension should only be used if extra values are "
                    f"available. Last requested dimensionality is already {dimensions[-1]}."
                )
            dimensions += (self.num_output,)
        if not _valid_dimensions(dimensions):
            raise ValueError(
                f"Max of given dimensions needs to match the number of outputs "
                f"in the encoder network. Got {dimensions} and expected a "
                f"maximum value of {self.num_output}.")

        if self.output_mode == self.Mode.OVERLAPPING:
            self.feature_ranges = tuple(
                slice(0, dimension) for dimension in dimensions)
        elif self.output_mode == self.Mode.SEPARATE:
            from_dimension = (0,) + dimensions
            self.feature_ranges = tuple(
                slice(i, j) for i, j in zip(from_dimension, dimensions))
        else:
            raise ValueError(
                f"Unknown mode: '{self.output_mode}', use one of {self.Mode._ALL}."
            )

    def forward(self, inputs):
        """Compute multiple embeddings for a single signal input.

        Args:
            inputs: The input tensor

        Returns:
            A tuple of tensors which are sliced according to `self.feature_ranges`
            if `renormalize` is set to true, each of the tensors will be normalized
            across the first (feature) dimension.

        TODO:
            - Cover this function with unit tests
        """
        output = self.module(inputs)
        outputs = (
            output[:, slice_features] for slice_features in self.feature_ranges)
        if self.renormalize:
            outputs = (self._norm(output) for output in outputs)
        return tuple(outputs)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import matplotlib
import numpy as np
import plotly.graph_objects as go
import pytest
from plotly.subplots import make_subplots

import cebra.integrations.plotly as cebra_plotly
import cebra.integrations.sklearn.cebra as cebra_sklearn_cebra


@pytest.mark.parametrize("cmap", ["viridis", "plasma", "inferno", "magma"])
def test_colorscale(cmap):
    cmap = matplotlib.cm.get_cmap(cmap)
    colorscale = cebra_plotly._convert_cmap2colorscale(cmap)
    assert isinstance(colorscale, list)


@pytest.mark.parametrize("output_dimension, idx_order", [(8, (2, 3, 4)),
                                                         (3, (0, 1, 2))])
def test_plot_embedding(output_dimension, idx_order):
    # example dataset
    X = np.random.uniform(0, 1, (1000, 50))
    y = np.random.uniform(0, 1, (1000,))

    # integration tests
    model = cebra_sklearn_cebra.CEBRA(max_iterations=10,
                                      batch_size=512,
                                      output_dimension=output_dimension)

    model.fit(X)
    embedding = model.transform(X)

    fig = cebra_plotly.plot_embedding_interactive(embedding=embedding,
                                                  embedding_labels=y)
    assert isinstance(fig, go.Figure)
    assert len(fig.data) == 1

    fig.layout = {}
    fig.data = []

    fig_subplots = make_subplots(
        rows=2,
        cols=2,
        specs=[
            [{
                "type": "scatter3d"
            }, {
                "type": "scatter3d"
            }],
            [{
                "type": "scatter3d"
            }, {
                "type": "scatter3d"
            }],
        ],
    )

    fig_subplots = cebra_plotly.plot_embedding_interactive(axis=fig_subplots,
                                                           embedding=embedding,
                                                           embedding_labels=y,
                                                           row=1,
                                                           col=1)

    fig_subplots.data = []
    fig_subplots.layout = {}


def test_discrete_with_legend():
    embedding = np.random.uniform(0, 1, (1000, 3))
    labels = np.random.randint(0, 10, (1000,))

    fig = cebra_plotly.plot_embedding_interactive(embedding,
                                                  labels,
                                                  discrete=True,
                                                  showlegend=True)

    assert len(fig._data_objs) == np.unique(labels).shape[0]
    assert isinstance(fig, go.Figure)


def test_continuous_no_legend():
    embedding = np.random.uniform(0, 1, (1000, 3))
    labels = np.random.uniform(0, 1, (1000,))

    fig = cebra_plotly.plot_embedding_interactive(embedding, labels)

    assert len(fig._data_objs) == 1

    assert isinstance(fig, go.Figure)


def test_continuous_with_legend_raises_error():
    embedding = np.random.uniform(0, 1, (1000, 3))
    labels = np.random.uniform(0, 1, (1000,))

    with pytest.raises(ValueError):
        cebra_plotly.plot_embedding_interactive(embedding,
                                                labels,
                                                discrete=False,
                                                showlegend=True)
